import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .AbstractScalarQuantizer import AbstractScalarQuantizer


@frozen
class ScalarQuantizer(AbstractScalarQuantizer):
    r"""
    General scalar quantizer. It is defined by a list of *levels*, $v_0, v_1, \ldots, v_{L-1}$, and a list of *thresholds*, $t_0, t_1, \ldots, t_L$, satisfying
    $$
        -\infty = t_0 < v_0 < t_1 < v_1 < \cdots < t_{L - 1} < v_{L - 1} < t_L = +\infty.
    $$
    Given an input $x \in \mathbb{R}$, the output of the quantizer is given by $y = v_i$ if and only if $t_i \leq x < t_{i+1}$, where $i \in [0:L)$. For more details, see <cite>Say06, Ch. 9</cite>.

    To invoke the quantizer, call the object giving the input signal as parameter (see example below).

    Attributes:
        levels (Array1D[float]): The quantizer levels $v_0, v_1, \ldots, v_{L-1}$. It should be a list floats of length $L$.

        thresholds (Array1D[float]): The quantizer finite thresholds $t_1, t_2, \ldots, t_{L-1}$. It should be a list of floats of length $L - 1$.

    Examples:
        The following example considers the $5$-level scalar quantizer whose characteristic (input × output) curve is depicted in the figure below.

        <figure markdown>
        ![Scalar quantizer example.](/figures/scalar_quantizer_5.svg)
        </figure>

        The levels are
        $$
            v_0 = -2, ~ v_1 = -1, ~ v_2 = 0, ~ v_3 = 1, ~ v_4 = 2,
        $$
        and the thresholds are
        $$
            t_0 = -\infty, ~ t_1 = -1.5, ~ t_2 = -0.3, ~ t_3 = 0.8, ~ t_4 = 1.4, ~ t_5 = \infty.
        $$

        >>> quantizer = komm.ScalarQuantizer(levels=[-2.0, -1.0, 0.0, 1.0, 2.0], thresholds=[-1.5, -0.3, 0.8, 1.4])
        >>> x = np.linspace(-2.5, 2.5, num=11)
        >>> y = quantizer(x)
        >>> np.vstack([x, y])
        array([[-2.5, -2. , -1.5, -1. , -0.5,  0. ,  0.5,  1. ,  1.5,  2. ,  2.5],
               [-2. , -2. , -1. , -1. , -1. ,  0. ,  0. ,  1. ,  2. ,  2. ,  2. ]])
    """

    levels: npt.NDArray[np.float64] = field(
        converter=np.asarray, repr=lambda x: x.tolist()
    )
    thresholds: npt.NDArray[np.float64] = field(
        converter=np.asarray, repr=lambda x: x.tolist()
    )

    def __attrs_post_init__(self) -> None:
        if self.thresholds.size != self.num_levels - 1:
            raise ValueError("length of 'thresholds' must be 'num_levels - 1'")

        interleaved = np.empty(2 * self.num_levels - 1, dtype=float)
        interleaved[0::2] = self.levels
        interleaved[1::2] = self.thresholds

        if not np.array_equal(np.unique(interleaved), interleaved):
            raise ValueError("invalid values for 'levels' and 'thresholds'")

    @property
    def num_levels(self) -> int:
        r"""
        The number of quantization levels $L$.
        """
        return self.levels.size

    def __call__(self, input_signal: npt.ArrayLike) -> npt.NDArray[np.float64]:
        tiled = np.tile(input_signal, reps=(self.thresholds.size, 1)).transpose()
        output_signal = self.levels[np.sum(tiled >= self.thresholds, axis=1)]
        return output_signal



import itertools as it
from functools import cache, cached_property

import numpy as np
import numpy.typing as npt
from attrs import frozen

from .BlockCode import BlockCode
from .matrices import reed_muller_generator_matrix


@frozen
class ReedMullerCode(BlockCode):
    r"""
    Reed–Muller code. It is a [linear block code](/ref/BlockCode) defined by two integers $\rho$ and $\mu$, which must satisfy $0 \leq \rho < \mu$. See references for more details. The resulting code is denoted by $\mathrm{RM}(\rho, \mu)$, and has the following parameters:

    - Length: $n = 2^{\mu}$
    - Dimension: $k = 1 + {\mu \choose 1} + \cdots + {\mu \choose \rho}$
    - Redundancy: $m = 1 + {\mu \choose 1} + \cdots + {\mu \choose \mu - \rho - 1}$
    - Minimum distance: $d = 2^{\mu - \rho}$

    For more details, see <cite>LC04, Sec. 4.3</cite>.

    Notes:
        - For $\rho = 0$ it reduces to a [repetition code](/ref/RepetitionCode).
        - For $\rho = 1$ it reduces to a lengthened [simplex code](/ref/SimplexCode).
        - For $\rho = \mu - 2$ it reduces to an extended [Hamming code](/ref/HammingCode).
        - For $\rho = \mu - 1$ it reduces to a [single parity check code](/ref/SingleParityCheckCode).

    Attributes:
        rho: The parameter $\rho$ of the code.
        mu: The parameter $\mu$ of the code.

    The parameters must satisfy $0 \leq \rho < \mu$.

    Examples:
        >>> code = komm.ReedMullerCode(1, 5)
        >>> (code.length, code.dimension, code.redundancy)
        (32, 6, 26)
        >>> code.generator_matrix  # doctest: +NORMALIZE_WHITESPACE
        array([[0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
               [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
               [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
               [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1],
               [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
               [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]])
        >>> code.minimum_distance()
        16
    """

    rho: int
    mu: int

    def __attrs_post_init__(self) -> None:
        if not 0 <= self.rho < self.mu:
            raise ValueError("'rho' and 'mu' must satisfy 0 <= rho < mu")

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.int_]:
        return reed_muller_generator_matrix(self.rho, self.mu)

    @cache
    def minimum_distance(self) -> int:
        return 2 ** (self.mu - self.rho)

    @property
    def default_decoder(self) -> str:
        return "reed"

    @classmethod
    def supported_decoders(cls) -> list[str]:
        return cls.__base__.supported_decoders() + ["reed", "weighted-reed"]  # type: ignore

    @cache
    def reed_partitions(self) -> list[npt.NDArray[np.int_]]:
        r"""
        The Reed partitions of the code. See <cite>LC04, Sec. 4.3</cite>.

        Examples:
            >>> code = komm.ReedMullerCode(2, 4)
            >>> reed_partitions = code.reed_partitions()
            >>> reed_partitions[1]
            array([[ 0,  1,  4,  5],
                   [ 2,  3,  6,  7],
                   [ 8,  9, 12, 13],
                   [10, 11, 14, 15]])
            >>> reed_partitions[8]
            array([[ 0,  4],
                   [ 1,  5],
                   [ 2,  6],
                   [ 3,  7],
                   [ 8, 12],
                   [ 9, 13],
                   [10, 14],
                   [11, 15]])
        """
        rho, mu = self.rho, self.mu
        reed_partitions: list[npt.NDArray[np.int_]] = []
        binary_vectors = [
            np.fliplr(np.array(list(it.product([0, 1], repeat=ell)), dtype=int))
            for ell in range(mu + 1)
        ]
        for ell in range(rho, -1, -1):
            for indices in it.combinations(range(mu), ell):
                setI = np.array(indices, dtype=int)
                setE = np.setdiff1d(np.arange(mu), indices, assume_unique=True)
                setS = np.dot(binary_vectors[ell], 2**setI)
                setQ = np.dot(binary_vectors[mu - ell], 2**setE)
                reed_partitions.append(setS[np.newaxis] + setQ[np.newaxis].T)
        return reed_partitions



from functools import cache, cached_property

import numpy as np
import numpy.typing as npt
from attrs import field, frozen
from tqdm import tqdm

from komm._error_control_block import BlockCode
from komm._util.bit_operations import int2binlist


@frozen(eq=False)
class Lexicode(BlockCode):
    r"""
    Lexicographic code (lexicode). For a given length $n$ and minimum distance $d$, it is the [linear block code](/ref/BlockCode) obtained by starting with the all-zero codeword and adding all binary $n$-tuples (in lexicographic order) that are at least at distance $d$ from all codewords already in the code.

    Attributes:
        n: The length $n$ of the code.
        d: The minimum distance $d$ of the code.

    For more details, see <cite>HP03, Sec. 2.11</cite>.

    Examples:
        >>> code = komm.Lexicode(7, 3)  # Hamming (7, 4)
        >>> (code.length, code.dimension, code.redundancy)
        (7, 4, 3)
        >>> code.generator_matrix
        array([[0, 0, 0, 0, 1, 1, 1],
               [0, 0, 1, 1, 0, 0, 1],
               [0, 1, 0, 1, 0, 1, 0],
               [1, 0, 0, 1, 0, 1, 1]])
        >>> code.minimum_distance()
        3
    """

    n: int
    d: int
    _codewords: list[int] = field(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        if not 1 <= self.d <= self.n:
            raise ValueError("'n' and 'd' must satisfy 1 <= d <= n")
        object.__setattr__(self, "_codewords", self._generate_codewords())

    def _generate_codewords(self) -> list[int]:
        codewords = [0]
        for i in tqdm(range(1, 2**self.n), desc="Generating lexicode", delay=2.5):
            # Reverse checking is way faster [Wikipedia].
            if all((c ^ i).bit_count() >= self.d for c in reversed(codewords)):
                codewords.append(i)
        return codewords

    @property
    def length(self) -> int:
        return self.n

    @cache
    def minimum_distance(self) -> int:
        return self.d

    @cached_property
    def generator_matrix(self) -> npt.NDArray[np.int_]:
        n = self.n
        codewords = self._codewords
        k = len(codewords).bit_length() - 1
        generator_matrix = np.empty((k, n), dtype=int)
        for i in range(k):
            generator_matrix[i] = list(reversed(int2binlist(codewords[1 << i], n)))
        return generator_matrix



import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .._util.information_theory import LogBase, entropy
from .._validation import is_pmf


@frozen
class DiscreteMemorylessSource:
    r"""
    Discrete memoryless source (DMS). It is defined by an *alphabet* $\mathcal{X}$ and a *probability mass function* (pmf) $p_X$. Here, for simplicity, the alphabet is always taken as $\mathcal{X} = \\{ 0, 1, \ldots, |\mathcal{X}| - 1 \\}$. The pmf $p_X$ gives the probability of the source emitting the symbol $X = x$.

    To invoke the source, call the object giving the number of symbols to be emitted as parameter (see example below).

    Attributes:
        pmf: The source probability mass function $p_X$. The element in position $x \in \mathcal{X}$ must be equal to $p_X(x)$.

    Examples:
        >>> np.random.seed(42)
        >>> dms = komm.DiscreteMemorylessSource([0.5, 0.4, 0.1])
        >>> dms(10)
        array([0, 2, 1, 1, 0, 0, 0, 1, 1, 1])
    """

    pmf: npt.NDArray[np.float64] = field(converter=np.asarray, validator=is_pmf)

    @property
    def cardinality(self) -> int:
        r"""
        The cardinality $|\mathcal{X}|$ of the source alphabet.
        """
        return self.pmf.size

    def entropy(self, base: LogBase = 2.0) -> float:
        r"""
        Returns the source entropy $\mathrm{H}(X)$. See [`komm.entropy`](/ref/entropy) for more details.

        Parameters:
            base (Optional[float | str]): See [`komm.entropy`](/ref/entropy). The default value is $2.0$.

        Examples:
            >>> dms = komm.DiscreteMemorylessSource([1/2, 1/4, 1/8, 1/8])
            >>> dms.entropy()
            np.float64(1.75)
            >>> dms.entropy(base=4)
            np.float64(0.875)
        """
        return entropy(self.pmf, base)

    def __call__(self, size: int) -> npt.NDArray[np.int_]:
        return np.random.choice(self.pmf.size, p=self.pmf, size=size)



from typing import Any, cast

import numpy as np
import numpy.typing as npt
from attr import field
from attrs import field, frozen

from .._util.information_theory import LogBase, binary_entropy
from .._validation import is_pmf, is_probability, validate_call
from .AbstractDiscreteMemorylessChannel import AbstractDiscreteMemorylessChannel


@frozen
class BinaryErasureChannel(AbstractDiscreteMemorylessChannel):
    r"""
    Binary erasure channel (BEC). It is a [discrete memoryless channel](/ref/DiscreteMemorylessChannel) with input alphabet $\mathcal{X} = \\{ 0, 1 \\}$ and output alphabet $\mathcal{Y} = \\{ 0, 1, 2 \\}$. The channel is characterized by a parameter $\epsilon$, called the *erasure probability*. With probability $1 - \epsilon$, the output symbol is identical to the input symbol, and with probability $\epsilon$, the output symbol is replaced by an erasure symbol (denoted by $2$). For more details, see <cite>CT06, Sec. 7.1.5</cite>.

    Attributes:
        erasure_probability (Optional[float]): The channel erasure probability $\epsilon$. Must satisfy $0 \leq \epsilon \leq 1$. Default value is `0.0`, which corresponds to a noiseless channel.

    Parameters: Input:
        input_sequence (Array1D[int]): The input sequence.

    Parameters: Output:
        output_sequence (Array1D[int]): The output sequence.

    Examples:
        >>> np.random.seed(1)
        >>> bec = komm.BinaryErasureChannel(0.1)
        >>> bec([1, 1, 1, 0, 0, 0, 1, 0, 1, 0])
        array([1, 1, 2, 0, 0, 2, 1, 0, 1, 0])
    """

    erasure_probability: float = field(default=0.0, validator=is_probability)

    @property
    def input_cardinality(self) -> int:
        return 2

    @property
    def output_cardinality(self) -> int:
        return 3

    @property
    def transition_matrix(self) -> npt.NDArray[np.float64]:
        r"""
        The transition probability matrix of the channel, given by
        $$
            p_{Y \mid X} =
            \begin{bmatrix}
                1 - \epsilon & 0 & \epsilon \\\\
                0 & 1 - \epsilon & \epsilon
            \end{bmatrix}.
        $$

        Examples:
            >>> bec = komm.BinaryErasureChannel(0.1)
            >>> bec.transition_matrix
            array([[0.9, 0. , 0.1],
                   [0. , 0.9, 0.1]])
        """
        epsilon = self.erasure_probability
        return np.array([[1 - epsilon, 0, epsilon], [0, 1 - epsilon, epsilon]])

    @validate_call(input_pmf=field(converter=np.asarray, validator=is_pmf))
    def mutual_information(
        self, input_pmf: npt.ArrayLike, base: LogBase = 2.0
    ) -> float:
        r"""
        Returns the mutual information $\mathrm{I}(X ; Y)$ between the input $X$ and the output $Y$ of the channel. It is given by
        $$
            \mathrm{I}(X ; Y) = (1 - \epsilon) \, \Hb(\pi),
        $$
        in bits, where $\pi = \Pr[X = 1]$, and $\Hb$ is the [binary entropy function](/ref/binary_entropy).

        **Parameters:**

        Same as the [corresponding method](/ref/DiscreteMemorylessChannel/#mutual_information) of the general class.

        Examples:
            >>> bec = komm.BinaryErasureChannel(0.1)
            >>> bec.mutual_information([0.45, 0.55]) # doctest: +NUMBER
            np.float64(0.8934970085890275)
        """
        input_pmf = cast(npt.NDArray[np.float64], input_pmf)
        epsilon = self.erasure_probability
        pi = input_pmf[1]
        return (1.0 - epsilon) * binary_entropy(pi) / np.log2(base)

    def capacity(self, base: LogBase = 2.0, **kwargs: Any) -> float:
        r"""
        Returns the channel capacity $C$. It is given by
        $$
            C = 1 - \epsilon,
        $$
        in bits.

        Examples:
            >>> bec = komm.BinaryErasureChannel(0.1)
            >>> bec.capacity()
            np.float64(0.9)
        """
        return (1.0 - self.erasure_probability) / np.log2(base)

    def __call__(self, input_sequence: npt.ArrayLike) -> npt.NDArray[np.int_]:
        epsilon = self.erasure_probability
        erasure_pattern = np.random.rand(np.size(input_sequence)) < epsilon
        output_sequence = np.copy(input_sequence)
        output_sequence[erasure_pattern] = 2
        return output_sequence



import numpy as np
import numpy.typing as npt
from attrs import frozen

from .VariableToFixedCode import VariableToFixedCode


@frozen
class VariableToFixedDecoder:
    r"""
    Decoder for [variable-to-fixed length codes](/ref/VariableToFixedCode).

    Attributes:
        code: The code to be considered.

    Parameters: Input:
        in0 (Array1D[int]): The sequence of symbols to be decoded. Must be a 1D-array with elements in $[0:T)$, where $T$ is the target cardinality of the code.

    Parameters: Output:
        out0 (Array1D[int]): The sequence of decoded symbols. It is a 1D-array with elements in $[0:S)$, where $S$ is the source cardinality of the code.

    Examples:
        >>> code = komm.VariableToFixedCode.from_sourcewords(2, [(0,0,0), (0,0,1), (0,1), (1,)])
        >>> decoder = komm.VariableToFixedDecoder(code)
        >>> decoder([0, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0])
        array([0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0])
    """

    code: VariableToFixedCode

    def __call__(self, in0: npt.ArrayLike) -> npt.NDArray[np.int_]:
        n, dec = self.code.target_block_size, self.code.dec_mapping
        out0 = np.concatenate([dec[tuple(s)] for s in np.reshape(in0, shape=(-1, n))])
        return out0



from functools import cache
from typing import Literal

import numpy as np
import numpy.typing as npt
from attrs import field, frozen

from .AbstractScalarQuantizer import AbstractScalarQuantizer


@frozen
class UniformQuantizer(AbstractScalarQuantizer):
    r"""
    Uniform scalar quantizer. It is a [scalar quantizer](/ref/ScalarQuantizer) in which the separation between levels is constant, $\Delta$, and the thresholds are the mid-point between adjacent levels. For more details, see <cite>Say06, Sec. 9.4</cite>.

    Attributes:
        num_levels: The number of quantization levels $L$. It must be greater than $1$.

        input_range: The range $(x_\mathrm{min}, x_\mathrm{max})$ of the input signal. The default is $(-1.0, 1.0)$.

        choice: The choice for the uniform quantizer. Must be either `'mid-riser'` or `'mid-tread'`. The default is `'mid-riser'`.

    Examples:
        >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(-1.0, 1.0), choice='mid-riser')
        >>> quantizer.levels
        array([-0.75, -0.25,  0.25,  0.75])
        >>> quantizer.thresholds
        array([-0.5,  0. ,  0.5])

        >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(-1.0, 1.0), choice='mid-tread')
        >>> quantizer.levels
        array([-1. , -0.5,  0. ,  0.5])
        >>> quantizer.thresholds
        array([-0.75, -0.25,  0.25])

        >>> quantizer = komm.UniformQuantizer(num_levels=4, input_range=(0.0, 1.0), choice='mid-tread')
        >>> quantizer.levels
        array([0.  , 0.25, 0.5 , 0.75])
        >>> quantizer.thresholds
        array([0.125, 0.375, 0.625])
    """

    _num_levels: int = field(alias="num_levels")
    input_range: tuple[float, float] = field(default=(-1.0, 1.0))
    choice: Literal["mid-riser", "mid-tread"] = field(default="mid-riser")

    def __attrs_post_init__(self) -> None:
        if self.num_levels < 2:
            raise ValueError("'num_levels' must be greater than 1")
        if self.choice not in ["mid-riser", "mid-tread"]:
            raise ValueError("'choice' must be either 'mid-riser' or 'mid-tread'")

    @property
    def num_levels(self) -> int:
        return self._num_levels

    @property
    @cache
    def quantization_step(self) -> float:
        r"""
        The quantization step $\Delta$.
        """
        x_min, x_max = self.input_range
        delta = (x_max - x_min) / self.num_levels
        return delta

    @property
    @cache
    def levels(self) -> npt.NDArray[np.float64]:
        r"""
        The quantizer levels $v_0, v_1, \ldots, v_{L-1}$.
        """
        num = self.num_levels
        x_min, x_max = self.input_range
        delta = (x_max - x_min) / num
        endpoint = (num % 2 == 0) if self.choice == "mid-riser" else (num % 2 == 1)
        min_level = x_min + (delta / 2) * endpoint
        max_level = x_max - (delta / 2) * endpoint
        return np.linspace(min_level, max_level, num=num, endpoint=endpoint)

    @property
    @cache
    def thresholds(self) -> npt.NDArray[np.float64]:
        r"""
        The quantizer finite thresholds $t_1, t_2, \ldots, t_{L-1}$.
        """
        return (self.levels + self.quantization_step / 2)[:-1]

    def __call__(self, input_signal: npt.ArrayLike) -> npt.NDArray[np.float64]:
        input_signal = np.array(input_signal, dtype=float, ndmin=1)
        delta = self.quantization_step
        if self.choice == "mid-riser":
            quantized = delta * (np.floor(input_signal / delta) + 0.5)
        else:
            quantized = delta * np.floor(input_signal / delta + 0.5)
        output_signal = np.clip(quantized, a_min=self.levels[0], a_max=self.levels[-1])
        return output_signal



import numpy as np
import numpy.typing as npt
from attrs import field, mutable

from .._types import ArrayIntLike
from .._util.bit_operations import pack, unpack
from .ConvolutionalCode import ConvolutionalCode


@mutable
class ConvolutionalStreamEncoder:
    r"""
    Convolutional stream encoder. Encode a bit stream using a given [convolutional code](/ref/ConvolutionalCode). The internal state of the encoder is maintained across each call.

    Attributes:
        convolutional_code: The convolutional code.
        state: The current state of the encoder. The default value is `0`.

    Parameters: Input:
        in0 (Array1D[int]): The bit sequence to be encoded.

    Parameters: Output:
        out0 (Array1D[int]): The encoded bit sequence.

    Examples:
        >>> convolutional_code = komm.ConvolutionalCode([[0o7, 0o5]])
        >>> convolutional_encoder = komm.ConvolutionalStreamEncoder(convolutional_code)
        >>> convolutional_encoder([1, 1, 1, 1])
        array([1, 1, 0, 1, 1, 0, 1, 0])
        >>> convolutional_encoder([1, 1, 1, 1])
        array([1, 0, 1, 0, 1, 0, 1, 0])
    """

    convolutional_code: ConvolutionalCode
    state: int = field(default=0)

    def __call__(self, in0: ArrayIntLike) -> npt.NDArray[np.int_]:
        n = self.convolutional_code.num_output_bits
        k = self.convolutional_code.num_input_bits
        fsm = self.convolutional_code.finite_state_machine()
        output_sequence, self.state = fsm.process(
            input_sequence=pack(in0, width=k),
            initial_state=self.state,
        )
        out0 = unpack(output_sequence, width=n)
        return out0



import numpy as np
import numpy.typing as npt

from ..registry import RegistryBlockDecoder
from ..SingleParityCheckCode import SingleParityCheckCode


def decode_wagner(
    code: SingleParityCheckCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    # See Costello, Forney: Channel Coding: The Road to Channel Capacity.
    r = np.asarray(r)
    v_hat = (r < 0).astype(int)
    if np.count_nonzero(v_hat) % 2 != 0:
        i = np.argmin(np.abs(r))
        v_hat[i] ^= 1
    return v_hat


RegistryBlockDecoder.register(
    "wagner",
    {
        "description": (
            "Wagner decoder. A soft-decision decoder for SingleParityCheck codes only."
        ),
        "decoder": decode_wagner,
        "type_in": "soft",
        "type_out": "hard",
        "target": "codeword",
    },
)



import numpy as np
import numpy.typing as npt
from attrs import frozen

from .FixedToVariableCode import FixedToVariableCode
from .util import parse_prefix_free


@frozen
class FixedToVariableDecoder:
    r"""
    Decoder for prefix-free [fixed-to-variable length codes](/ref/FixedToVariableCode).

    Attributes:
        code: The code to be considered, which must be a prefix-free code (that is, no codeword is a prefix of another codeword).

    Parameters: Input:
        in0 (Array1D[int]): The sequence of symbols to be decoded. Must be a 1D-array with elements in $[0:T)$, where $T$ is the target cardinality of the code.

    Parameters: Output:
        out0 (Array1D[int]): The sequence of decoded symbols. It is a 1D-array with elements in $[0:S)$, where $S$ is the source cardinality of the code.

    Examples:
        >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1,0), (1,1)])
        >>> decoder = komm.FixedToVariableDecoder(code)
        >>> decoder([1, 0, 0, 1, 0, 0, 1, 1, 0])
        array([1, 0, 1, 0, 2, 0])

        >>> code = komm.FixedToVariableCode.from_codewords(2, [(0,), (1,0), (1,1), (1,1,0)])
        >>> decoder = komm.FixedToVariableDecoder(code)
        Traceback (most recent call last):
        ...
        ValueError: code is not prefix-free
    """

    code: FixedToVariableCode

    def __attrs_post_init__(self) -> None:
        if not self.code.is_prefix_free():
            raise ValueError("code is not prefix-free")

    def __call__(self, in0: npt.ArrayLike) -> npt.NDArray[np.int_]:
        in0 = np.asarray(in0)
        out0 = parse_prefix_free(in0, self.code.inv_enc_mapping)
        return out0



from .bcjr import *
from .berlekamp import *
from .exhaustive_search import *
from .majority_logic_repetition_code import *
from .meggitt import *
from .reed import *
from .syndrome_table import *
from .viterbi import *
from .wagner import *

__all__ = [
    "decode_bcjr",
    "decode_berlekamp",
    "decode_exhaustive_search_hard",
    "decode_exhaustive_search_soft",
    "decode_majority_logic_repetition_code",
    "decode_meggitt",
    "decode_reed",
    "decode_syndrome_table",
    "decode_viterbi_hard",
    "decode_viterbi_soft",
    "decode_wagner",
]



from typing import Literal

import numpy.typing as npt

from .constellations import constellation_apsk
from .labelings import get_labeling
from .Modulation import Modulation


class APSKModulation(Modulation):
    r"""
    Amplitude- and phase-shift keying (APSK) modulation. It is a complex [modulation scheme](/ref/Modulation) in which the constellation is the union (concatenation) of component [PSK](/ref/PSKModulation) constellations, called *rings*. More precisely, consider $K$ rings $\mathbf{X}_k$, for $k \in [0 : K)$, where the $k$-th ring has order $M_k$, amplitude $A_k$, and phase offset $\phi_k$. The $i$-th constellation symbol of the $k$-th ring is given by
    $$
        x\_{k,i} = A_k \exp \left( \mathrm{j} \frac{2 \pi i}{M_k} \right) \exp(\mathrm{j} \phi_k),
        \quad k \in [0 : K),
        \quad i \in [0 : M_k).
    $$
    The resulting APSK constellation is therefore given by
    $$
        \mathbf{X} = \begin{bmatrix}
            \mathbf{X}_0 \\\\
            \vdots \\\\
            \mathbf{X}\_{K-1}
        \end{bmatrix},
    $$
    which has order $M = M_0 + M_1 + \cdots + M\_{K-1}$. The order $M_k$ of each ring need not be a power of $2$; however, the order $M$ of the constructed APSK modulation must be. The APSK constellation is depicted below for $(M_0, M_1) = (8, 8)$ with $(A_0, A_1) = (A, 2A)$ and $(\phi_0, \phi_1) = (0, \pi/8)$; and for $(M_0, M_1) = (4, 12)$ with $(A_0, A_1) = (\sqrt{2}A, 3A)$ and $(\phi_0, \phi_1) = (\pi/4, 0)$.

    <div class="centered" markdown>
      <span>
        ![(8,8)-APSK constellation.](/figures/apsk_8_8.svg)
      </span>
      <span>
        ![(4,12)-APSK constellation.](/figures/apsk_4_12.svg)
      </span>
    </div>

    Parameters:
        orders: A $K$-tuple with the orders $M_k$ of each ring, for $k \in [0 : K)$. The sum $M_0 + M_1 + \cdots + M_{K-1}$ must be a power of $2$.

        amplitudes: A $K$-tuple with the amplitudes $A_k$ of each ring, for $k \in [0 : K)$.

        phase_offsets: A $K$-tuple with the phase offsets $\phi_k$ of each ring, for $k \in [0 : K)$. If specified as a single float $\phi$, then it is assumed that $\phi_k = \phi$ for all $k \in [0 : K)$. The default value is `0.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be equal to `'natural'`. The default value is `'natural'`.

    Examples:
        >>> apsk = komm.APSKModulation(orders=(8, 8), amplitudes=(1.0, 2.0), phase_offsets=(0.0, np.pi/8))
        >>> apsk.constellation.round(4)
        array([ 1.    +0.j    ,  0.7071+0.7071j,  0.    +1.j    , -0.7071+0.7071j,
               -1.    +0.j    , -0.7071-0.7071j, -0.    -1.j    ,  0.7071-0.7071j,
                1.8478+0.7654j,  0.7654+1.8478j, -0.7654+1.8478j, -1.8478+0.7654j,
               -1.8478-0.7654j, -0.7654-1.8478j,  0.7654-1.8478j,  1.8478-0.7654j])

        >>> apsk = komm.APSKModulation(orders=(4, 12), amplitudes=(np.sqrt(2), 3.0), phase_offsets=(np.pi/4, 0.0))
        >>> apsk.constellation.round(4)
        array([ 1.    +1.j    , -1.    +1.j    , -1.    -1.j    ,  1.    -1.j    ,
                3.    +0.j    ,  2.5981+1.5j   ,  1.5   +2.5981j,  0.    +3.j    ,
               -1.5   +2.5981j, -2.5981+1.5j   , -3.    +0.j    , -2.5981-1.5j   ,
               -1.5   -2.5981j, -0.    -3.j    ,  1.5   -2.5981j,  2.5981-1.5j   ])
    """

    def __init__(
        self,
        orders: tuple[int, ...],
        amplitudes: tuple[float, ...],
        phase_offsets: float | tuple[float, ...] = 0.0,
        labeling: Literal["natural"] | npt.ArrayLike = "natural",
    ) -> None:
        _phase_offsets = (
            phase_offsets
            if isinstance(phase_offsets, tuple)
            else (phase_offsets,) * len(orders)
        )
        super().__init__(
            constellation=constellation_apsk(orders, amplitudes, _phase_offsets),
            labeling=get_labeling(labeling, ("natural",), sum(orders)),
        )
        self.orders = orders
        self.amplitudes = amplitudes
        self.phase_offsets = phase_offsets
        self.labeling_parameter = labeling

    def __repr__(self) -> str:
        args = ", ".join([
            f"orders={self.orders}",
            f"amplitudes={self.amplitudes}",
            f"phase_offsets={self.phase_offsets}",
            f"labeling='{self.labeling_parameter}'",
        ])
        return f"{self.__class__.__name__}({args})"



import math
from typing import Iterable, Literal

import numpy.typing as npt

from .constellations import constellation_qam
from .labelings import get_labeling
from .Modulation import Modulation


class QAModulation(Modulation):
    r"""
    Quadrature-amplitude modulation (QAM). It is a complex [modulation scheme](/ref/Modulation) in which the constellation is given as a Cartesian product of two [PAM](/ref/PAModulation) constellations, namely, the *in-phase constellation*, and the *quadrature constellation*. More precisely, the $i$-th constellation symbol is given by
    $$
        \begin{aligned}
            x_i = \left[ A_\mathrm{I} \left( 2i_\mathrm{I} - M_\mathrm{I} + 1 \right) + \mathrm{j} A_\mathrm{Q} \left( 2i_\mathrm{Q} - M_\mathrm{Q} + 1 \right) \right] \exp(\mathrm{j}\phi), \quad
                &  i \in [0 : M), \\\\
                & i_\mathrm{I} = i \bmod M_\mathrm{I}, \\\\
                & i_\mathrm{Q} = \lfloor i / M_\mathrm{I} \rfloor,
        \end{aligned}
    $$
    where $M_\mathrm{I}$ and $M_\mathrm{Q}$ are the *orders* (powers of $2$), and $A_\mathrm{I}$ and $A_\mathrm{Q}$ are the *base amplitudes* of the in-phase and quadrature constellations, respectively. Also, $\phi$ is the *phase offset*. The order of the resulting complex-valued constellation is $M = M_\mathrm{I} M_\mathrm{Q}$, a power of $2$. The QAM constellation is depicted below for $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 4)$ with ($A_\mathrm{I}, A_\mathrm{Q}) = (A, A)$; and for $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 2)$ with $(A_\mathrm{I}, A_\mathrm{Q}) = (A, 2A)$; in both cases, $\phi = 0$.

    <div class="centered" markdown>
      <span>
        ![16-QAM constellation.](/figures/qam_16.svg)
      </span>
      <span>
        ![8-QAM constellation.](/figures/qam_8.svg)
      </span>
    </div>

    Parameters:
        orders: A tuple $(M_\mathrm{I}, M_\mathrm{Q})$ with the orders of the in-phase and quadrature constellations, respectively; both $M_\mathrm{I}$ and $M_\mathrm{Q}$ must be powers of $2$. If specified as a single integer $M$, then it is assumed that $M_\mathrm{I} = M_\mathrm{Q} = \sqrt{M}$; in this case, $M$ must be an square power of $2$.

        base_amplitudes: A tuple $(A_\mathrm{I}, A_\mathrm{Q})$ with the base amplitudes of the in-phase and quadrature constellations, respectively. If specified as a single float $A$, then it is assumed that $A_\mathrm{I} = A_\mathrm{Q} = A$. The default value is $1.0$.

        phase_offset: The phase offset $\phi$ of the constellation. The default value is `0.0`.

        labeling: The binary labeling of the modulation. Can be specified either as a 2D-array of integers (see [base class](/ref/Modulation) for details), or as a string. In the latter case, the string must be either `'natural_2d'` or `'reflected_2d'`. The default value is `'reflected_2d'`, corresponding to the Gray labeling.

    Examples:
        The square $16$-QAM modulation with $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 4)$ and $(A_\mathrm{I}, A_\mathrm{Q}) = (1, 1)$, and Gray labeling is depicted below.

        <figure markdown>
          ![16-QAM modulation with Gray labeling.](/figures/qam_16_gray.svg)
        </figure>

        >>> qam = komm.QAModulation(16)
        >>> qam.constellation  # doctest: +NORMALIZE_WHITESPACE
        array([-3.-3.j, -1.-3.j,  1.-3.j,  3.-3.j,
               -3.-1.j, -1.-1.j,  1.-1.j,  3.-1.j,
               -3.+1.j, -1.+1.j,  1.+1.j,  3.+1.j,
               -3.+3.j, -1.+3.j,  1.+3.j,  3.+3.j])
        >>> qam.labeling   # doctest: +NORMALIZE_WHITESPACE
        array([[0, 0, 0, 0], [1, 0, 0, 0], [1, 1, 0, 0], [0, 1, 0, 0],
               [0, 0, 1, 0], [1, 0, 1, 0], [1, 1, 1, 0], [0, 1, 1, 0],
               [0, 0, 1, 1], [1, 0, 1, 1], [1, 1, 1, 1], [0, 1, 1, 1],
               [0, 0, 0, 1], [1, 0, 0, 1], [1, 1, 0, 1], [0, 1, 0, 1]])
        >>> qam.modulate([0, 0, 1, 1, 0, 0, 1, 0])
        array([-3.+1.j, -3.-1.j])

        The rectangular $8$-QAM modulation with $(M_\mathrm{I}, M_\mathrm{Q}) = (4, 2)$ and $(A_\mathrm{I}, A_\mathrm{Q}) = (1, 2)$, and Gray labeling is depicted below.

        <figure markdown>
            ![8-QAM modulation with Gray labeling.](/figures/qam_8_gray.svg)
        </figure>

        >>> qam = komm.QAModulation(orders=(4, 2), base_amplitudes=(1.0, 2.0))
        >>> qam.constellation  # doctest: +NORMALIZE_WHITESPACE
        array([-3.-2.j, -1.-2.j,  1.-2.j,  3.-2.j,
               -3.+2.j, -1.+2.j,  1.+2.j,  3.+2.j])
        >>> qam.labeling  # doctest: +NORMALIZE_WHITESPACE
        array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0],
               [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]])
        >>> qam.modulate([0, 0, 1, 1, 0, 0, 1, 0, 1])
        array([-3.+2.j, -1.-2.j, -1.+2.j])
    """

    def __init__(
        self,
        orders: tuple[int, int] | int,
        base_amplitudes: tuple[float, float] | float = 1.0,
        phase_offset: float = 0.0,
        labeling: (
            Literal["natural_2d", "reflected_2d"] | npt.ArrayLike
        ) = "reflected_2d",
    ) -> None:
        _orders = (
            orders
            if isinstance(orders, Iterable)
            else (math.isqrt(orders), math.isqrt(orders))
        )
        _base_amplitudes = (
            base_amplitudes
            if isinstance(base_amplitudes, Iterable)
            else (base_amplitudes, base_amplitudes)
        )
        super().__init__(
            constellation=constellation_qam(_orders, _base_amplitudes, phase_offset),
            labeling=get_labeling(labeling, ("natural_2d", "reflected_2d"), _orders),
        )
        self.orders = orders
        self.base_amplitudes = base_amplitudes
        self.phase_offset = phase_offset
        self.labeling_parameter = labeling

    def __repr__(self):
        args = ", ".join([
            f"orders={self.orders}",
            f"base_amplitudes={self.base_amplitudes}",
            f"phase_offset={self.phase_offset}",
            f"labeling='{self.labeling_parameter}'",
        ])
        return f"{self.__class__.__name__}({args})"



import numpy as np
import numpy.typing as npt

from ..BlockCode import BlockCode
from ..registry import RegistryBlockDecoder


def decode_exhaustive_search_hard(
    code: BlockCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    codewords = code.codewords()
    metrics = np.count_nonzero(r != codewords, axis=1)
    v_hat = codewords[np.argmin(metrics)]
    return v_hat


RegistryBlockDecoder.register(
    "exhaustive-search-hard",
    {
        "description": (
            "Exhaustive search (hard-decision). Minimum Hamming distance decoder"
        ),
        "decoder": decode_exhaustive_search_hard,
        "type_in": "hard",
        "type_out": "hard",
        "target": "codeword",
    },
)


def decode_exhaustive_search_soft(
    code: BlockCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    codewords = code.codewords()
    metrics = np.dot(codewords, r)
    v_hat = codewords[np.argmin(metrics)]
    return v_hat


RegistryBlockDecoder.register(
    "exhaustive-search-soft",
    {
        "description": (
            "Exhaustive search (soft-decision). Minimum Euclidean distance decoder"
        ),
        "decoder": decode_exhaustive_search_soft,
        "type_in": "soft",
        "type_out": "hard",
        "target": "codeword",
    },
)



import numpy as np
import numpy.typing as npt
from attrs import frozen

from .FixedToVariableCode import FixedToVariableCode


@frozen
class FixedToVariableEncoder:
    r"""
    Encoder for [fixed-to-variable length codes](/ref/FixedToVariableCode).

    Attributes:
        code: The code to be considered.

    Parameters: Input:
        in0 (Array1D[int]): The sequence of symbols to be encoded. Must be a 1D-array with elements in $[0:S)$, where $S$ is the source cardinality of the code.

    Parameters: Output:
        out0 (Array1D[int]): The sequence of encoded symbols. It is a 1D-array with elements in $[0:T)$, where $T$ is the target cardinality of the code.

    Examples:
        >>> code = komm.FixedToVariableCode.from_codewords(3, [(0,), (1,0), (1,1)])
        >>> encoder = komm.FixedToVariableEncoder(code)
        >>> encoder([1, 0, 1, 0, 2, 0])
        array([1, 0, 0, 1, 0, 0, 1, 1, 0])
    """

    code: FixedToVariableCode

    def __call__(self, in0: npt.ArrayLike) -> npt.NDArray[np.int_]:
        k, enc = self.code.source_block_size, self.code.enc_mapping
        out0 = np.concatenate([enc[tuple(s)] for s in np.reshape(in0, shape=(-1, k))])
        return out0



from typing import Literal, cast

import numpy as np
import numpy.typing as npt
from attrs import field, validators

from .._validation import is_pmf, validate_call
from .FixedToVariableCode import FixedToVariableCode
from .util import huffman_algorithm


@validate_call(
    pmf=field(converter=np.asarray, validator=is_pmf),
    source_block_size=field(validator=validators.ge(1)),
    policy=field(validator=validators.in_(["high", "low"])),
)
def HuffmanCode(
    pmf: npt.ArrayLike,
    source_block_size: int = 1,
    policy: Literal["high", "low"] = "high",
) -> FixedToVariableCode:
    r"""
    Binary Huffman code. It is an optimal (minimal expected rate) [fixed-to-variable length code](/ref/FixedToVariableCode) for a given probability mass function. For more details, see <cite>Say06, Sec. 3.2</cite>.

    Parameters:
        pmf: The probability mass function of the source.
        source_block_size: The source block size $k$. The default value is $k = 1$.
        policy: The policy to be used when constructing the code. It must be either `'high'` (move combined symbols as high as possible) or `'low'` (move combined symbols as low as possible). The default value is `'high'`.

    Examples:
        >>> pmf = [0.7, 0.15, 0.15]

        >>> code = komm.HuffmanCode(pmf)
        >>> code.enc_mapping  # doctest: +NORMALIZE_WHITESPACE
        {(0,): (0,),
         (1,): (1, 1),
         (2,): (1, 0)}
        >>> code.rate(pmf)  # doctest: +NUMBER
        np.float64(1.3)

        >>> code = komm.HuffmanCode(pmf, 2)
        >>> code.enc_mapping  # doctest: +NORMALIZE_WHITESPACE
        {(0, 0): (1,),
         (0, 1): (0, 0, 0, 0),
         (0, 2): (0, 1, 1),
         (1, 0): (0, 1, 0),
         (1, 1): (0, 0, 0, 1, 1, 1),
         (1, 2): (0, 0, 0, 1, 1, 0),
         (2, 0): (0, 0, 1),
         (2, 1): (0, 0, 0, 1, 0, 1),
         (2, 2): (0, 0, 0, 1, 0, 0)}
        >>> code.rate(pmf)  # doctest: +NUMBER
        np.float64(1.1975)
    """
    pmf = cast(npt.NDArray[np.float64], pmf)
    return FixedToVariableCode.from_codewords(
        source_cardinality=pmf.size,
        codewords=huffman_algorithm(pmf, source_block_size, policy),
    )



import numpy as np
import numpy.typing as npt

from ..registry import RegistryBlockDecoder
from ..RepetitionCode import RepetitionCode


def decode_majority_logic_repetition_code(
    code: RepetitionCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    u_hat = np.array([np.argmax(np.bincount(r))])
    return u_hat


RegistryBlockDecoder.register(
    "majority-logic-repetition-code",
    {
        "description": (
            "Majority-logic decoder. A hard-decision decoder for Repetition codes only."
        ),
        "decoder": decode_majority_logic_repetition_code,
        "type_in": "hard",
        "type_out": "hard",
        "target": "message",
    },
)



from typing import Callable

import numpy as np
import numpy.typing as npt

from .ScalarQuantizer import ScalarQuantizer
from .util import lloyd_max_quantizer


def LloydMaxQuantizer(
    input_pdf: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]],
    input_range: tuple[float, float],
    num_levels: int,
) -> ScalarQuantizer:
    r"""
    Lloyd–Max scalar quantizer. It is a [scalar quantizer](/ref/ScalarQuantizer) that minimizes the mean-squared error (MSE) between the input signal $X$ and its quantized version. For more details, see <cite>Say06, Sec. 9.6.1</cite>.

    Parameters:
        input_pdf: The probability density function $f_X(x)$ of the input signal.

        input_range: The range $(x_\mathrm{min}, x_\mathrm{max})$ of the input signal.

        num_levels: The number $L$ of quantization levels. It must be greater than $1$.

    Examples:
        >>> uniform_pdf = lambda x: 1/8 * (np.abs(x) <= 4)
        >>> quantizer = komm.LloydMaxQuantizer(
        ...     input_pdf=uniform_pdf,
        ...     input_range=(-4, 4),
        ...     num_levels=8,
        ... )
        >>> quantizer.levels
        array([-3.5, -2.5, -1.5, -0.5,  0.5,  1.5,  2.5,  3.5])
        >>> quantizer.thresholds
        array([-3., -2., -1.,  0.,  1.,  2.,  3.])

        >>> gaussian_pdf = lambda x: 1/np.sqrt(2*np.pi) * np.exp(-x**2/2)
        >>> quantizer = komm.LloydMaxQuantizer(
        ...     input_pdf=gaussian_pdf,
        ...     input_range=(-5, 5),
        ...     num_levels=8,
        ... )
        >>> quantizer.levels.round(3)
        array([-2.152, -1.344, -0.756, -0.245,  0.245,  0.756,  1.344,  2.152])
        >>> quantizer.thresholds.round(3)  # doctest: +SKIP
        array([-1.748, -1.05 , -0.501, -0.   ,  0.501,  1.05 ,  1.748])
    """
    levels, thresholds = lloyd_max_quantizer(
        input_pdf,
        num_levels,
        input_range,
        points_per_interval=1000,
        max_iter=100,
    )
    return ScalarQuantizer(levels=levels, thresholds=thresholds)



import numpy as np
import numpy.typing as npt

from ..ReedMullerCode import ReedMullerCode
from ..registry import RegistryBlockDecoder


def decode_reed(
    code: ReedMullerCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    # See [LC04, p. 105–114, 439–440].
    u_hat = np.empty(code.dimension, dtype=int)
    bx = np.copy(r)
    for i, partition in enumerate(code.reed_partitions()):
        checksums = np.count_nonzero(bx[partition], axis=1) % 2
        u_hat[i] = np.count_nonzero(checksums) > len(checksums) // 2
        bx ^= u_hat[i] * code.generator_matrix[i]
    return u_hat


RegistryBlockDecoder.register(
    "reed",
    {
        "description": (
            "Reed decoding algorithm for Reed–Muller codes. It's a majority-logic"
            " decoding algorithm."
        ),
        "decoder": decode_reed,
        "type_in": "hard",
        "type_out": "hard",
        "target": "message",
    },
)


def decode_weighted_reed(
    code: ReedMullerCode,
    r: npt.ArrayLike,
) -> npt.NDArray[np.int_]:
    # See [LC04, p. 440–442].
    r = np.asarray(r)
    u_hat = np.empty(code.dimension, dtype=int)
    bx = (r < 0).astype(int)
    for i, partition in enumerate(code.reed_partitions()):
        checksums = np.count_nonzero(bx[partition], axis=1) % 2
        min_reliability = np.min(np.abs(r[partition]), axis=1)
        decision_var = np.dot(1 - 2 * checksums, min_reliability)
        u_hat[i] = decision_var < 0
        bx ^= u_hat[i] * code.generator_matrix[i]
    return u_hat


RegistryBlockDecoder.register(
    "weighted-reed",
    {
        "description": "Weighted Reed decoding algorithm for Reed–Muller codes.",
        "decoder": decode_weighted_reed,
        "type_in": "soft",
        "type_out": "hard",
        "target": "message",
    },
)
